/* Driver for Interlan Ethernet Card */
/* B. Doster - First pass 3/9/88 */

#define	TIMER	20000	/* Timeout on transmissions */

#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "enet.h"
#include "iface.h"
#include "il.h"
#include "timer.h"
#include "arp.h"
#include "trace.h"
#include "pc.h"
#include "iptrace.h"

static void il_rcv_reset(register unsigned);
static void il_rcv_pkt(struct il *, unsigned int);

extern int il0vec();
int (*ilvec[])() = {il0vec};
struct il il[IL_MAX];		/* Per-controller info */
unsigned nil = 0;


/* Initialize interface */
int
il_init(interface,bufsize)
struct interface *interface;
unsigned bufsize;	/* Maximum size of receive queue in PACKETS */
{
	register struct il *ilp;
	register unsigned base;
	void eaudit();
	unsigned short getcs();
	char i_state;
	int16 dev;

	dev = interface->dev;
	ilp = &il[dev];
	base = ilp->base;
	ilp->rcvmax = bufsize;
	ilp->iface = interface;

	i_state = disable();	/* No distractions from the receiver */

	/* Hold up EDLC RESET while the board is configured */
	outportb(EDLC_RESET(base),RS_RESET);

	/* Hardware reset of Interlan Board */
	outportb(IE_RESET(base),0);

	/* Set interrupt vector */
	setirq(ilp->vec,ilvec[dev]);

	/* Clear mask (enable interrupt) in 8259 interrupt controller */
	maskon(ilp->vec);

	/* Shouldn't be any interrupts from here on out */
	restore(i_state);

	outportb(EDLC_XMODE(base),XMD_LBC) ;

	/* Read and Set Ethernet Board Address */
	if(interface->hwaddr == NULLCHAR)
		interface->hwaddr = malloc(EADDR_LEN);
	getiladdr(base,interface->hwaddr);
	setiladdr(base,interface->hwaddr);
	if(memcmp(interface->hwaddr,ether_bdcst,EADDR_LEN) == 0){
		printf("IL address PROM contains broadcast address!!\n");
		return -1;
	}

	/* Only enable Transmit-type interrupts while Transmitting */
	outportb(EDLC_XMASK(base),0) ;

	/* Establish generic Transmit mode */
	outportb(EDLC_XMODE(base),XMD_IG_PAR | XMD_T_MODE | XMD_LBC) ;

	/* Clear any startup related Transmit interrupts */
	outportb(EDLC_XCLR(base),0xFF) ;

	/* Establish generic Receive mode */
	outportb(EDLC_RMODE(base),RMD_BROADCAST) ;

	/* Reset Rcv State to allow receives */
	il_rcv_reset(base) ;

	/* Finally un-reset the EDLC */
	outportb(EDLC_RESET(base),0);

	return 0;
}

/* Send an IP datagram on Ethernet */
il_send(bp,interface,gateway,precedence,delay,throughput,reliability)
struct mbuf *bp;		/* Buffer to send */
struct interface *interface;	/* Pointer to interface control block */
int32 gateway;			/* IP address of next hop */
char precedence;
char delay;
char throughput;
char reliability;
{
	char *egate,*res_arp();

	egate = res_arp(interface,ARP_ETHER,gateway,bp);
	if(egate != NULLCHAR)
		(*interface->output)(interface,egate,interface->hwaddr,IP_TYPE,bp);
}

/* Send a packet with Ethernet header */
il_output(interface,dest,source,type,bp)
struct interface *interface;		/* Pointer to interface control block */
char dest[];		/* Destination Ethernet address */
char source[];		/* Source Ethernet address */
int16 type;		/* Type field */
struct mbuf *bp;	/* Data field */
{
	struct ether ep;
	struct mbuf *hdr;

	memcpy(ep.dest,dest,EADDR_LEN);
	memcpy(ep.source,source,EADDR_LEN);
	ep.type = type;
	hdr = htonether(&ep);
	hdr->next = bp;
	(*interface->raw)(interface,hdr);
}

/* Send raw packet (caller provides header) */
il_raw(interface,bp)
struct interface *interface;	/* Pointer to interface control block */
struct mbuf *bp;		/* Data field */
{
	register struct il *ilp;
	register unsigned base;
	register int i;
	char i_state;
	short size;

	ilp = &il[interface->dev];
	base = ilp->base;

	dump(interface,IF_TRACE_OUT,TRACE_ETHER,bp);

	size = len_mbuf(bp);
	/* Pad the size out to the minimum, if necessary,
	 * with junk from the last packet (nice security hole here)
	 */
	if(size < RUNT)
		size = RUNT;
	size = (size+1) & ~1;	/* round size up to next even number */

	/* Wait for transmitter ready, if necessary. IE_XMTBSY is valid
	 * only in the transmit mode, hence the initial check.
	 */
	if (inportb(IE_ISTAT(base)) & IS_EN_XMT) { /* on-going xmt */
		for (i=TIMER;(inportb(IE_ISTAT(base)) & IS_EN_XMT) && i != 0; i--)
			;
		if (i == 0) {
			ilp->estats.timeout++ ;
			free_p(bp) ;
			return ;
		}
	}

	ilp->size = size;
	i_state = disable();	/* No distractions from the receiver */

	/* Disable the Receiver */
	outportb(EDLC_RMASK(base), 0) ;	/* Mask all Receive Interrupts */
	outportb(IE_MMODE(base), 0) ;	/* Put Xmt Buffer on System Bus */
	outportb(EDLC_RCLR(base), 0xFF) ;/* Clear out any pending Rcv Ints */

	/* Point GP at beginning of packet */
	outportw(IE_GP(base), BFRSIZ-size) ;

	/* Actually load each piece with a fast assembler routine */
	while(bp != NULLBUF){
		outbuf(IE_XBUF(base),bp->data,bp->cnt) ;
		bp = free_mbuf(bp) ;
	}

	/* Rewrite where packet starts */
	outportw(IE_GP(base),BFRSIZ-size) ;

	/* Flip Xmt Buffer to EDLC Bus so it can be transmitted */
	outportb(IE_MMODE(base), MM_MUX) ;

	/* Begin transmission. */
	outportb(IE_MMODE(base), MM_EN_XMT | MM_MUX) ;

	/* Cause interrupt after completion (or if something fails) */
	outportb(EDLC_XMASK(base), XM_TPOK | XM_RCVD | XM_UFLW |
				   XM_COLL | XM_COLL16		) ;

	restore(i_state);
}

/* Ethernet interrupt handler */
int
ilint(dev)
int dev;
{
	register struct il *ilp;
	register unsigned base;
	char intstat;
	int16 size;
	char stat;

	ilp = &il[dev];
	base = il[dev].base;
	ilp->estats.intrpt++;
	ip_trace.totint++;
	intstat = inportb(IE_ISTAT(base)) ;

	/* DMA Complete Interrupt.  We don't use DMA, but just in case ... */
	if (!(intstat & IS_DMA_INT)) {
		ilp->estats.dmadone++;
		outportb(IE_DMA_RST(base), 0) ; /* Reset DMA Interrupt */
	}

	/* Inc Badint once for each type of Interrupt (Xmt & Rcv) */
	switch (intstat & (IS_X_INT | IS_R_INT)) {
		case IS_X_INT | IS_R_INT: /* neither Xmt, nor Rcv */
			ip_trace.badint++;
			break;
		case IS_X_INT:		  /* Rcv Interrupt */
		case IS_R_INT:		  /* Xmt Interrupt */
			ip_trace.badint++;
			break;
		default:		  /* Xmt & Rcv Interrupt */
			ip_trace.badint += 2;
			break;
	}

	/* Transmit Complete/Fail Interrupt */
	if (!(intstat & IS_X_INT)) {
		stat = inportb(EDLC_XSTAT(base)) ;
		outportb(EDLC_XMASK(base), 0) ;		/* Disable Xmt IRQ's */
		outportb(EDLC_XCLR(base), 0xFF) ;	/* clr all Xmt IRQ's */

		if (stat & XS_16COLL) /* 16th jam.  give up on sending it */
			ilp->estats.jam16++;
		else if (stat & XS_COLL) {
			/* Crank counter back to beginning and restart xmt */
			ilp->estats.jam++;

			/* Point GP at beginning of packet */
			outportw(IE_GP(base), BFRSIZ - ilp->size) ;

			/* De-assert MM_EN_RCV bit Just In Case */
			outportb(IE_MMODE(base), 0) ;

			/* Flip Xmt Buffer to EDLC Bus and restart xmt */
			outportb(IE_MMODE(base), MM_EN_XMT | MM_MUX) ;

			/* Interrupt for all Transmit errors along with TPOK */
			outportb(EDLC_XMASK(base), XM_TPOK | XM_RCVD | XM_UFLW |
						   XM_COLL | XM_COLL16	) ;

			/* Wait for it to complete again */
			return ;
		} else if (stat & XS_TPOK) {
			ip_trace.badint--;
			ip_trace.xmtint++;
			ilp->estats.xmit++;  /* count it as a good packet */
		}
	}

	/* Receive Packet Interrupt */
	if (!(intstat & IS_R_INT)) {
		stat = inportb(EDLC_RSTAT(base)) ;
		if ((stat & RS_VALID_BITS) == RS_PKT_OK)
			il_rcv_pkt(ilp,base) ;
		else {
			if (stat & RS_RST_PKT)
				ilp->estats.rstpkt++ ;
			if (stat & RS_RUNT)
				ilp->estats.runt++ ;
			if (stat & RS_OFLW)
				ilp->estats.over++ ;
			ilp->estats.bad++ ;
		}
	}

	/* Re-prime for another receive */
	il_rcv_reset(base) ;
}
/* Receive a "Known To Be Good" packet */
static
void
il_rcv_pkt(ilp,base)
struct il *ilp ;
unsigned int base ;
{
	struct mbuf *bp;
	int16 size ;	/* Size of received packet */

	/* Clear the interrupt */
	outportb(EDLC_RCLR(base), 0xFF) ;

	size = inportw(IE_RCNT(base)) ;

	/* Flip Rcv Buffer onto the system bus */
	outportb(IE_MMODE(base), MM_MUX) ;

	/* Point to beginning of Rcv'd packet */
	outportw(IE_GP(base), 0) ;

	/* Allocate mbuf and copy the packet into it */
	if (size > GIANT)
		ilp->estats.bad++ ;
	else if (ilp->rcvcnt >= ilp->rcvmax)
		ilp->estats.drop++ ;
	else if ((bp = alloc_mbuf(size)) == NULLBUF)
		ilp->estats.nomem++ ;
	else {
		ilp->estats.recv++ ;
		inbuf(IE_RBUF(base),bp->data,size);
		bp->cnt = size;
		ilp->rcvcnt++;
		ip_trace.badint--;
		ip_trace.queint++;
		enqueue(&ilp->rcvq,bp); /* Put it on the receive queue */
	}
}
/* Prime Interlan card for another Receive */
static
void
il_rcv_reset(base)
register unsigned base;
{
	/* Rcv packet at start of Boards buffer */
	outportw(IE_GP(base),0) ;

	/* Clear any remaining Interrupt conditions */
	outportb(EDLC_RCLR(base),0xFF) ;

	/* Set MUX to allow EDLC to access Rcv Buffer */
	outportb(IE_MMODE(base),0) ;

	/* Enable Receive of Normal and Broadcast Packets only. */
	outportb(EDLC_RMODE(base),RMD_BROADCAST) ;

	/* Enable Receive Interrupts */
	outportb(IE_MMODE(base),MM_EN_RCV) ;

	/* Unmask *all* Receive related interrupts */
	outportb(EDLC_RMASK(base),0xFF) ;
}
/* Process any incoming Ethernet packets on the receive queue */
int
il_recv(interface)
struct interface *interface;
{
	void arp_input(),ip_route();
	struct il *ilp;
	struct mbuf *bp;
	struct ether hdr;
	char multicast;

	ilp = &il[interface->dev];
	if((bp = dequeue(&ilp->rcvq)) == NULLBUF)
		return;

	ilp->rcvcnt--;
	ip_trace.queint--;
	ip_trace.rcvint++;

	dump(interface,IF_TRACE_IN,TRACE_ETHER,bp);

	/* Remove Ethernet header and kick packet upstairs */
	ntohether(&hdr,&bp);
	if(hdr.dest[0] & 1)
		multicast = 1;
	else
		multicast = 0;
	switch(hdr.type){
	case ARP_TYPE:
		arp_input(interface,bp);
		break;
	case IP_TYPE:
		ip_route(bp,multicast);
		break;
	default:
		ip_trace.totother++;
		free_p(bp);
		break;
	}
}
/* Read Ethernet address from controller PROM */
static
getiladdr(base,cp)
register unsigned base;
register char *cp;
{
	register int i;
	register unsigned char u;

	for (i=0;i<EADDR_LEN;i++){
		outportw(IE_GP(base),i);
		*cp++ = inportb(IE_SAPROM(base));
	}
	for(i=6;i<9;i++){
		outportw(IE_GP(base),i);
		u = (unsigned char)inportb(IE_SAPROM(base));
		switch (i) {
			case 6: if (u != 0x00) printf("PROM 6 == %X\n",u) ;
				break ;
			case 7: if (u != 0x55) printf("PROM 7 == %X\n",u) ;
				break ;
			case 8: if (u != 0xAA) printf("PROM 8 == %X\n",u) ;
				break ;
		}
	}
}
/* Set Ethernet address on controller */
static
setiladdr(base,cp)
register unsigned base;
register char *cp;
{
	register int i;

	for (i=0;i<EADDR_LEN;i++)
		outportb(EDLC_ADDR(base)+i,*cp++);
}
/* Shut down the Ethernet controller */
il_stop(interface)
struct interface *interface;
{
	register unsigned base;
	char i_state;
	int16 dev;

	dev = interface->dev;
	base = il[dev].base;

	/* put card in held-RESET state */
	outportb(IE_MMODE(base),0) ;
	outportb(EDLC_RESET(base),RS_RESET) ;

	i_state = disable();	/* No distractions from the receiver */

	/* Set mask (disable interrupt) in 8259 interrupt controller */
	maskoff(il[dev].vec);

	restore(i_state);
}
/* Attach a Interlan NI5010 Ethernet controller to the system
 * argv[0]: hardware type, must be "NI5010"
 * argv[1]: I/O address, e.g., "0x300"
 * argv[2]: vector, e.g., "3"
 * argv[3]: mode, must be "arpa"
 * argv[4]: interface label, e.g., "il0"
 * argv[5]: maximum number of packets allowed on receive queue, e.g., "5"
 * argv[6]: maximum transmission unit, bytes, e.g., "1500"
 */
il_attach(argc,argv)
int argc;
char *argv[];
{
	register struct interface *if_il;
	extern struct interface *ifaces;
	unsigned dev;
	int il_init();
	int il_send();
	int il_recv();
	int il_stop();
	int pether(),gaether();

	arp_init(ARP_ETHER,EADDR_LEN,IP_TYPE,ARP_TYPE,ether_bdcst,pether,gaether);
	if(nil >= IL_MAX){
		printf("Too many Ethernet controllers\n");
		return -1;
	}
	dev = nil++;
	if((if_il = (struct interface *)calloc(1,sizeof(struct interface))) == NULLIF
	 ||(if_il->name = malloc((unsigned)strlen(argv[4])+1)) == NULLCHAR){
		printf("il_attach: no memory!\n");
		return -1;
	}
	strcpy(if_il->name,argv[4]);
	if_il->mtu = atoi(argv[6]);
	if_il->send = il_send;
	if_il->output = il_output;
	if_il->raw = il_raw;
	if_il->recv = il_recv;
	if_il->stop = il_stop;
	if_il->dev = dev;

	il[dev].base = htoi(argv[1]);
	il[dev].vec = htoi(argv[2]);

	if(strcmp(argv[3],"arpa") != 0){
		printf("Mode %s unknown for interface %s\n",
			argv[3],argv[4]);
		free(if_il->name);
		free((char *)if_il);
		return -1;
	}
	/* Initialize device */
	if(il_init(if_il,(unsigned)atoi(argv[5])) != 0){
		free(if_il->name);
		free((char *)if_il);
		return -1;
	}
	if_il->next = ifaces;
	ifaces = if_il;

	return 0;
}
